
/*
	clock.cpp

	This program animates the motion of a pendulum clock.
*/


#include <stdio.h>
#include <stdlib.h>
#include <string>
#include <math.h>

// The path to the glut head file may differ on your system
#include <GL/glut.h>

#include "st.h"


// Program globals ----------------------------------


// ST Scene graph object
STScene clockScene;

// Pointers to nodes in the scene graph
STSceneNode* escapementGear;
STSceneNode* pendulumGroup;

STTimer clockTimer;

// scales time (useful for debugging)
float timeSpeedup = 1.0f;

const int framesPerSecond = 30;

int windowSizeX;
int windowSizeY;

int savedFrames = 0;



// This method saves the current state of the framebuffer
// as an image on disk.
static void writeFrame(const char* name, int frameNumber) {

	char filename[80];
	sprintf(filename, "%s%04d.png", name, frameNumber);

	/*
		CS148 Todo:

		Use the STImage class to write the contents of the window
		out to the file 'filename'

	*/

	printf("Wrote frame %d as %s\n", frameNumber, filename);
}



// 'timerTick' is the event handler for GLUT timer events
// The position of the pendulum and clock
// gears should be updated here
static void timerTick(int value) {

	// get the amount of wall clock time passed since the timer was
	// started (it was started in the setup() function)
	float totalTime = clockTimer.GetElapsedMillis();

	printf("Elapsed time = %.3fsec\n", totalTime / 1000.0f);

	float theta, x, y, z;

	/*
		CS148 Todo:

		Currently the implementation of timerTick just alters the rotation
		applied to the shapes in the scene.  This needs to be changed to:

		 * Rotate the escapement gear at a rate of 1 tooth per second.
		 * Rotate the other gears in the gear trane at the appropriate rates
		 * Move the pendulum in a realistic way
	*/

	// Change the rotation transform on the escapement gear
	STTransform xform = escapementGear->GetTransform();
	xform.GetRotate(theta, x, y, z);
	xform.SetRotate(theta - timeSpeedup * .5f, 0.0f, 0.0f, 1.0f);
	escapementGear->SetTransform( xform );

	// Change the rotatation transform applied to the parent node
	// the parent of the red rectangle and the blue circle
	xform = pendulumGroup->GetTransform();
	xform.GetRotate(theta, x, y, z);
	xform.SetRotate( theta + timeSpeedup * .2f, 0.0f, 0.0f, 1.0f );
	pendulumGroup->SetTransform( xform );

	glutPostRedisplay();

	// schedule new timer event
	glutTimerFunc( (int)(1000.0f / framesPerSecond), timerTick, 0);
}


// This method is invoked on application start.  Construct the scene
// by adding all the asserts (shapes) into the scene graph here
static void setup() {

	// Get the root node of the scene graph from our scene object
	STSceneNode* root = clockScene.GetRoot();
	
	STSceneNode* node;

	STColor red(1.0f, 0.0f, 0.0f);
	STColor darkYellow(1.0f, 0.9f, .1f);
	STColor blue(.1f, .2f, .8f);

	/*
		CS148 Todo:

		The code below adds and positions three shapes into the scene.

		This needs to be changes to add all the pendulum clock shapes
		into the scene and set up a hierarchy of transforms that can be
		modified each time step to produce a realistic animation

	*/

	// --- escapement gear ----------------------------------------

	// Create an escapement gear shape, set its color to yellow
	STShape* shape = new STShape();
	shape->Load("escapement_gear_60.shp");
	shape->SetColor(darkYellow);

	// Now add the shape to the scene.  The result of this operation is
	// the creation of a new node in the scene graph (a child of the
	// root node).  The transform associated with this node can be modified
	// to move the shape on screen.
	escapementGear = root->PlaceObject(shape);


	// --- rotating rectangle -------------------------------------------

	// We'll now give a simple example of setting up a hierarchy of nodes
	// in the scene graph.  First we'll create a node that contains no
	// shapes.  In each time step we will rotate this node. 
	pendulumGroup = root->CreateNode();

	// Add a rectangle shape as a child node of the 'pendulum group' node.
	// The center the rectangle is at the origin in the shape's coordinate
	// space.  We need to apply a translation transform to the rectangle
	// so that the end of the shape is at the origin
	shape = STShape::CreateRect(.05f, 1.0f);
	shape->SetColor( red );
	node = pendulumGroup->PlaceObject(shape);
	node->SetTransform(STTransform::Translate(0.0f, -0.5f, 0.0f));
	
	// Add a circle shape to the scene as a child node of the 'pendulum group'
	// node.  Translate it so that is lies a quarter of the way along the rect
	shape = STShape::CreateCircle(.1f);
	shape->SetColor( blue );
	node = pendulumGroup->PlaceObject(shape);
	node->SetTransform(STTransform::Translate(0.0f, -0.25f, 0.0f));

	// start a timer to track how much wall clock time has passed
	clockTimer.Reset();

	// schedule first GLUT timer event
	glutTimerFunc( (int)(1000.0f / framesPerSecond), timerTick, 0);

}



static void display() {

	glClearColor(.1f, .1f, .1f, 1.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	glOrtho(-1.0, 1.0f, -1.0f, 1.0f, -.5f, .5f);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	// draw the scene
	clockScene.Draw();

	glutSwapBuffers();

}


// Keyboard event handler
static void keyboard(unsigned char key, int x, int y) {
  
	switch (key) {

		case char(27):
		case 'q':
			exit(0);

		case '+':
		case '=':
			timeSpeedup *= 2.0f;
			printf("Time is now scaled by %.2fX\n", timeSpeedup);
			break;

		case '-':
		case '_':
			timeSpeedup /= 2.0f;
			printf("Time is now scaled by %.2fX\n", timeSpeedup);
			break;

		case 's':
			// Save current state of framebuffer to disk
			writeFrame("clock", savedFrames);
			savedFrames++;
			break;
	}
}


// Reshape just records the new size of the window. 
void reshape(int w, int h) {
	windowSizeX = w;
	windowSizeY = h;
}


int main(int argc, char** argv) {
	
	glutInit(&argc, argv);
	glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
	glutInitWindowPosition(20,20);
	glutInitWindowSize(650, 650);
	glutCreateWindow("Pendulum Clock: CS148 Assignment 2");

	glEnable(GL_DEPTH_TEST);

	setup();

	glutDisplayFunc(display);
	glutKeyboardFunc(keyboard);
	glutReshapeFunc(reshape);
	glutMainLoop();	

	return 0;
}

