
#include <stdio.h>
#include <string>

#include "stgl.h"
#include "STImage.h"

extern "C" {
#include <jpeglib.h>    // libjpeg header
}

#include <png.h>		// libpng header
#include <setjmp.h>     // include of setjmp.h *MUST* come after the include ofpng.h 





/**
* Converts the 'ch' to upper case
*/
static char
STToUpper(char ch) {
	if ( ch >= 'a' && ch <= 'z')
		return (char)((int)ch - 32);
	else
		return ch;
}


STImage::STImage()
: mHeight(-1), mWidth(-1), mPixels(NULL) {
	
}


STImage::~STImage() {
	if (mPixels)
		delete [] mPixels;
}


/**
* Returns height of the image in pixels
*/
int
STImage::GetHeight() const {
	return mHeight;
}

/**
* Returns width of the image in pixels
*/
int
STImage::GetWidth() const {
	return mWidth;
}

/**
* Fills in image pixel data using a width by height region of the OpenGL
* framebuffer beginning at pixel (startx, starty).  
* This operation will destroy all previous pixel data in the STImage object
* and result in a resize of the image dimensions to be width by height
*/
void
STImage::Read(int startx, int starty, int width, int height) {
	
	if (mPixels)
		delete [] mPixels;

	mWidth = width;
	mHeight = height;

	mPixels = new STPixel[width * height];

	glReadPixels(startx, starty, width, height, GL_RGBA, GL_UNSIGNED_BYTE, (GLvoid*)mPixels );
}


/**
* Parse filename to obtain the file extension.  We'll use this
* to determine which file format the image is in
*/
static std::string
getExtension(const char* filename)
{

	std::string name = std::string(filename);
	size_t pos = name.find_last_of(".");

	std::string extension = name.substr(pos+1, name.size()-pos);
	
	for (size_t i=0; i<extension.size(); i++)
		extension[i] = STToUpper(extension[i]);

	return extension;
}

/**
* Loads an image off disk.  This method does not do any real work, it
* just calls a format specific method based on the image format (format
* determined by file extension)
*/ 
STStatus 
STImage::Load(const char* filename) {


	// Determine how to load the file from the file's extension
	// The ST graphics library currently supports files of type
	// PPM, PNG, and JPG
	
	if (filename == NULL)
		return ST_ERROR;

	std::string ext = getExtension( filename );

	if (ext.compare("PNG") == 0)
		return LoadPNG(filename);
	else if (ext.compare("JPG") == 0)
		return LoadJPG(filename);
	else if (ext.compare("PPM") == 0)
		return LoadPPM(filename);
	else {
		fprintf(stderr, "Unknown image file type.\n");
		return ST_ERROR;
	} 

}

/**
* Fills in image data from the contents of a PNG file via the libpng API
*/
STStatus
STImage::LoadPNG(const char* filename) {

	char pngHeader[8];

	//printf("Loading PNG.\n");

	FILE* imgFile = fopen(filename, "rb");
	
	if (!imgFile) {
		fprintf(stderr, "Could not open '%s'.\n", filename);
		return ST_ERROR;
	}

	// Read the first 8 bytes from the file and validate that the
	// file is a valid png file
	fread(pngHeader, 1, 8, imgFile);
	
	if (png_sig_cmp( (png_byte*)pngHeader, 0, 8)) {
		fprintf(stderr, "Could not open '%s'. Unexpected format for png file.\n", filename);
		fclose(imgFile);
		return ST_ERROR;
	}

	// The following code initializes the png struct, which serves as an opaque handle
	// to the png image, and the png info struct, which stores meta-information about the
	// image file that can be queried once the data is read

	png_structp pngPtr;
	png_infop infoPtr;
	
	// main png struct (opaque handle)
	pngPtr = png_create_read_struct(PNG_LIBPNG_VER_STRING, (png_voidp)NULL, NULL, NULL);
	
	if (!pngPtr) {
		fclose(imgFile);
		return ST_ERROR;
	}

	// info struct (user accessible)
	infoPtr = png_create_info_struct(pngPtr);

	if (!infoPtr) {
		png_destroy_read_struct(&pngPtr, (png_infopp)NULL, (png_infopp)NULL);
		fclose(imgFile);
		return ST_ERROR;
	}

	// jump pointer.  Failure during read inside libpng will cause 
	// control to jump here (simple fail out)
	if (setjmp(png_jmpbuf(pngPtr))) {
		fprintf(stderr, "Could not open '%s'.  Internal error in libpng.\n", filename);
		png_destroy_read_struct(&pngPtr, &infoPtr, (png_infopp)NULL);
		fclose(imgFile);
		return ST_ERROR;
	}

	png_init_io(pngPtr, imgFile);
	png_set_sig_bytes(pngPtr, 8);

	// The following code used the libpng high level interface
	// to read data from the file into the array of STPixel structs
	// representing an image in this class.  Format conversion into 8-bit
	// RGBA as currently required by this class, requires an additional
	// buffer copy.  Using the libpng **low-level interface** this buffer
	// copy and the associated buffer double allocation could be avoided.
	// The code would be complicated slightly with additional libpng
	// transform calls, so I will leave this for later if performance or
	// memory footprint becomes an issue.

	// These transforms get us very close to 8 bit per channel RGBA, but
	// not quite (won't extend from 1 or 3 components to 4.  Must use low
	// level interface to do this.
	int transforms =			
			PNG_TRANSFORM_STRIP_ALPHA |
			PNG_TRANSFORM_PACKING |
			PNG_TRANSFORM_PACKSWAP |
			PNG_TRANSFORM_EXPAND |
			PNG_TRANSFORM_SHIFT;

	// Tell libpng to load the image into memory
	png_read_png(pngPtr, infoPtr, transforms, NULL);

	// libpng exposes the data as an array of arrays of bytes
	png_bytepp pngRowPointers = png_get_rows(pngPtr, infoPtr);

	mWidth = png_get_image_width(pngPtr, infoPtr);
	mHeight = png_get_image_height(pngPtr, infoPtr);	
	mPixels = new STPixel[ mWidth * mHeight ];

	// printf("PNG image size is: %dx%d\n", mWidth, mHeight);

	int numChannels = png_get_channels(pngPtr, infoPtr);

	// Now copy the data read by libpng into the STImage class'
	// contiguous allocation of pixel data.  In doing so, explode
	// single channel images or images without an alpha channel
	// into full RGBA pixels.  
	//
	// We also need to flip the order of the rows of data.  Data in
	// the png file begins with the topmost row of the image.  The
	// STImage class stores data bottom row first to be consistent
	// with OpenGL pixel formats
	for (int i=0; i<mHeight; i++) {

		png_bytep srcPtr = pngRowPointers[mHeight-i-1]; 
		STPixel* curPixel = &mPixels[ i*mWidth ];

		for (int j=0; j<mWidth; j++) {

			switch (numChannels) {

			// monochrome, no alpha
			case 1:
				curPixel->r = curPixel->g = curPixel->b = *srcPtr++;
				curPixel->a = 255;
				break;

			// monochrome, with alpha
			case 2:
				curPixel->r = curPixel->g = curPixel->b = *srcPtr++;
				curPixel->a = *srcPtr++;
				break;

			// RGB, no alpha
			case 3:
				curPixel->r = *srcPtr++;
				curPixel->g = *srcPtr++;
				curPixel->b = *srcPtr++;
				curPixel->a = 255;
				break;
				
			// RGBA
			case 4:
				curPixel->r = *srcPtr++;
				curPixel->g = *srcPtr++;
				curPixel->b = *srcPtr++;
				curPixel->a = *srcPtr++;
				break;
			}

			curPixel++;
		}
	}

	// cleanup
	png_read_end(pngPtr, NULL);
	png_destroy_read_struct(&pngPtr, &infoPtr, (png_infopp)NULL);
	fclose(imgFile);


	return ST_OK;
}


struct STJpegErrorMgr {
  jpeg_error_mgr  pub;	     
  jmp_buf         setjmpBuf;	
};


/**
* Custom error handling routine for use with libjpeg
*/
METHODDEF(void)
STJpegErrorExit(j_common_ptr cinfo)
{

  STJpegErrorMgr* myerr = (STJpegErrorMgr*) cinfo->err;

  // Display the error message.
  (*cinfo->err->output_message) (cinfo);

  // Return control to the setjmp point
  longjmp(myerr->setjmpBuf, 1);
}


/**
* Fills in image data from the contents of a JPG file via the libjpeg API
*/
STStatus
STImage::LoadJPG(const char* filename) {

	// printf("Loading JPG.\n");

// bug: will remain commented out until problems with building
// libjpeg using visual studio are resolved.
#ifndef _WIN32

	jpeg_decompress_struct cinfo;
	STJpegErrorMgr jerr;

	FILE* imgFile = fopen(filename, "rb");
	
	if (!imgFile) {
		fprintf(stderr, "Could not open '%s'.\n", filename);
		return ST_ERROR;
	}

	cinfo.err = jpeg_std_error(&jerr.pub);
	jerr.pub.error_exit = STJpegErrorExit;
	
	if (setjmp(jerr.setjmpBuf)) {
		jpeg_destroy_decompress(&cinfo);
		fclose(imgFile);
		return ST_ERROR;
	}

	jpeg_create_decompress(&cinfo);
	jpeg_stdio_src(&cinfo, imgFile);
	jpeg_read_header(&cinfo, TRUE);
	jpeg_start_decompress(&cinfo);

	int rowStride = cinfo.output_width * cinfo.output_components;
  
	mWidth = cinfo.output_width;
	mHeight = cinfo.output_height;
	mPixels = new STPixel[ mWidth * mHeight ];
		
	// temporary buffer to hold the decompressed data from the JPEG
	// file before it is stuck into the mPixels array
	JSAMPARRAY rowBuffer = (*cinfo.mem->alloc_sarray)((j_common_ptr) &cinfo, JPOOL_IMAGE, rowStride, 1);	

	
	// load all rows of pixels
	while (cinfo.output_scanline < cinfo.output_height) {

		STPixel* curPixel = &mPixels[ mWidth * (mHeight-cinfo.output_scanline-1) ];

		jpeg_read_scanlines(&cinfo, rowBuffer, 1);

		unsigned char* buf = rowBuffer[0];
		
		if (cinfo.output_components == 3) {
		
			for (int i=0; i<mWidth; i++) {
				curPixel->r = *buf++;
				curPixel->g = *buf++;
				curPixel->b = *buf++;
				curPixel->a = 255;
				curPixel++;
			}

		} else {

			for (int i=0; i<mWidth; i++) {
				curPixel->r = curPixel->g = curPixel->b = *buf++;
				curPixel->a = 255;
				curPixel++;
			}
		}
	}

	// cleanup
	jpeg_finish_decompress(&cinfo);
	jpeg_destroy_decompress(&cinfo);
	fclose(imgFile);

	if (jerr.pub.num_warnings > 0) {
		printf("Note: libjpeg produced warnings when reading '%s'.\n", filename);
	}

#endif

	return ST_OK;

}

/**
* Pulls the next line of a PPM file out of the file stream and 
* strips out any comments.
*/
static int
PPMNextLine(FILE* ppmFile, char* line) {
	
	if (fgets(line, 80, ppmFile) != NULL ) {
	
		char* commentChar = strchr(line, '#');
		if (commentChar)
			*commentChar = '\0';
			return 1;
		
	} else {
		return 0;
	}
}

/**
* Fills in image data from the contents of a PPM file
*/
STStatus
STImage::LoadPPM(const char* filename) {
 
	//printf("Loading PPM.\n");

	// PPM format guarantees no line in the file is longer
	// than 70 characters
	char line[80];

	FILE* imgFile = fopen(filename, "r");

	if (!imgFile) {
		fprintf(stderr, "Could not open file '%s'.\n", filename);
		return ST_ERROR;
	}

	// Read the PPM header, it should begin with the characters 'P3'.
	// If it doesn't, complain about an invalid format
	fgets(line, 3, imgFile);

	if (strcmp(line, "P3") != 0) {
		fprintf(stderr, "Invalid PPM file format.\n");
		return ST_ERROR;
	}

	int val;

	int pos = 0;
	int header[3];

	while (pos < 3 && PPMNextLine(imgFile, line)) {
	
		// process lines of the PPM file until we've read
		// the image's width, height, and max pixel value

		char* tok = strtok(line, " \t\n");

		while (tok) {      
			sscanf(tok, "%d", &val);      
			//printf("Scanned: %d\n", val);
			header[pos++] = val;      
			tok = strtok(NULL, " \t\n");
		} 
	}

	mWidth = header[0];
	mHeight = header[1];
	int maxVal = header[2];

	//printf("width=%d height=%d maxval=%d\n", mWidth, mHeight, maxVal);

	int pixelsToRead = mWidth * mHeight;
	int curComponent = 0;
	

	// allocate storage for pixels
	mPixels = new STPixel[ mWidth * mHeight ];

	int pixelValues[3];

	pos = 0;

	while ( pos < (mWidth*mHeight) && PPMNextLine(imgFile, line)) {

		char* tok = strtok(line, " \t\n");

		while (tok) {

			sscanf(tok, "%d", &val); 
			//printf("Scanned: %d\n", val);
			pixelValues[curComponent] = val;
			
			if (curComponent == 2) {	
		
				// done with this pixel, store the pixel and move to next
				
				mPixels[pos].r = pixelValues[0] * 255 / maxVal;
				mPixels[pos].g = pixelValues[1] * 255 / maxVal;
				mPixels[pos].b = pixelValues[2] * 255 / maxVal;
				mPixels[pos].a = 255;
				
				curComponent = 0;
				pos++;
		 
			} else
				curComponent++;

			tok = strtok(NULL, " \t\n");
		} 
	}
	
	fclose(imgFile);


	return ST_OK;
}

/**
* Saves an image to disk.  This method does not do any real work, it
* just calls a format specific method based on the image format
* (format determined by file extension)
*/ 
STStatus
STImage::Save(const char* filename) {

	if (filename == NULL || mPixels == NULL)
		return ST_ERROR;
	
	std::string ext = getExtension( filename );
	
	if (ext.compare("PNG") == 0)
		return SavePNG(filename);
	else if (ext.compare("JPG") == 0)
		return SaveJPG(filename);
	else if (ext.compare("PPM") == 0 )
		return SavePPM(filename);
	else {
		fprintf(stderr, "Unknown image file type.\n");
		return ST_ERROR;
	} 
}

/**
* Creates a PNG file from the pixel contents of the STImage.  Makes calls
* to libpng
*/
STStatus
STImage::SavePNG(const char* filename) {

	// printf("Saving PNG.\n")

	FILE* imgFile = fopen(filename, "wb");
	
	if (!imgFile) {
		fprintf(stderr, "Could not open '%s'.\n", filename);
		return ST_ERROR;
	}
	
	png_structp pngPtr;
	png_infop infoPtr;

	// main png struct (opaque handle)
	pngPtr = png_create_write_struct(PNG_LIBPNG_VER_STRING, (png_voidp)NULL, NULL, NULL);
	
	if (!pngPtr) {
		fclose(imgFile);
		return ST_ERROR;
	}

	// info struct (user accessible)
	infoPtr = png_create_info_struct(pngPtr);

	if (!infoPtr) {
		png_destroy_write_struct(&pngPtr, (png_infopp)NULL);
		fclose(imgFile);
		return ST_ERROR;
	}

	// jump pointer.  Failure during read inside libpng will cause 
	// control to jump here (simple fail out)
	if (setjmp(png_jmpbuf(pngPtr))) {
		fprintf(stderr, "Could not write '%s'.  Internal error in libpng.\n", filename);
		png_destroy_write_struct(&pngPtr, &infoPtr);
		fclose(imgFile);
		return ST_ERROR;
	}

	png_init_io(pngPtr, imgFile);

	png_set_IHDR(pngPtr, infoPtr, mWidth, mHeight, 8,
		PNG_COLOR_TYPE_RGB_ALPHA,
		PNG_INTERLACE_NONE,
		PNG_COMPRESSION_TYPE_DEFAULT,
		PNG_FILTER_TYPE_DEFAULT);

	// allocate a set of row pointers that points directly into the existing
	// array of STPixels
	png_bytepp rowPointers = (png_bytepp)png_malloc( pngPtr, png_sizeof(png_bytep) * mHeight );

	for (int i=0; i<mHeight; i++)
		rowPointers[i] = (png_bytep)(mPixels + (mHeight-i-1)*mWidth);

	png_set_rows(pngPtr, infoPtr, rowPointers);

	// write the image data to the specified file
	png_write_png(pngPtr, infoPtr, PNG_TRANSFORM_IDENTITY, NULL);

	png_free( pngPtr, rowPointers ) ;

	// cleanup
	png_write_end(pngPtr, NULL);
	png_destroy_write_struct( &pngPtr, &infoPtr) ;
	fclose(imgFile);

	return ST_OK;
}


/**
* Creates a JPG file from the pixel contents of the STImage.  Makes calls
* to libjpg
*/

STStatus
STImage::SaveJPG(const char* filename) {

	//printf("Saving JPG.\n");

// bug: will remain commented out until problems with building
// libjpeg using visual studio are resolved.
#ifndef _WIN32

	FILE* imgFile = fopen(filename, "w");

	if (!imgFile) {
		fprintf(stderr, "Could not open '%s' for writing.\n", filename);
		return ST_ERROR;
	}

	jpeg_compress_struct cinfo;
	STJpegErrorMgr jerr;

	cinfo.err = jpeg_std_error(&jerr.pub);
	jerr.pub.error_exit = STJpegErrorExit;
	
	if (setjmp(jerr.setjmpBuf)) {
		jpeg_destroy_compress(&cinfo);
		fclose(imgFile);
		return ST_ERROR;
	}

	jpeg_create_compress(&cinfo);
	jpeg_stdio_dest(&cinfo, imgFile);

	cinfo.image_width = mWidth; 	
	cinfo.image_height = mHeight;
	cinfo.input_components = 3;		
	cinfo.in_color_space = JCS_RGB; 	

	jpeg_set_defaults(&cinfo);
	jpeg_set_quality(&cinfo, 90, TRUE);
	jpeg_start_compress(&cinfo, TRUE);

	int rowStride = mWidth * 3;

	JSAMPARRAY buffer = (*cinfo.mem->alloc_sarray)((j_common_ptr) &cinfo, JPOOL_IMAGE, rowStride, 1);	

	while (cinfo.next_scanline < cinfo.image_height) {
		
		STPixel* curPixel = &mPixels[ mWidth * (mHeight-cinfo.next_scanline-1) ];

		JSAMPLE* buf = buffer[0];
		for (int i=0; i<mWidth; i++) {
			*buf++ = curPixel->r;
			*buf++ = curPixel->g;
			*buf++ = curPixel->b;
			curPixel++;
		}

		jpeg_write_scanlines(&cinfo, buffer, 1);
	}

	jpeg_finish_compress(&cinfo);
	fclose(imgFile);
	jpeg_destroy_compress(&cinfo);

#endif

	return ST_OK;
}



/**
* Creates a PPM file from the pixel contents of the STImage
*/
STStatus
STImage::SavePPM(const char* filename) {

	//printf("Saving PPM.\n");

	FILE* imgFile = fopen(filename, "w");

	if (!imgFile) {
		fprintf(stderr, "Could not open '%s' for writing.\n", filename);
		return ST_ERROR;
	}

	fprintf(imgFile, "P3\n");
	fprintf(imgFile, "%d %d\n", mWidth, mHeight);
	fprintf(imgFile, "255\n");
	
	int totalPixels = mWidth * mHeight;
	for (int i=0; i<totalPixels; i++)
		fprintf(imgFile,"%d %d %d\n", mPixels[i].r, mPixels[i].g, mPixels[i].b);	
	
	fclose(imgFile);

	return ST_OK;
}

/**
* Draws the image to the screen with the bottom left corner at (0.0, 0.0)
* in object space
*/
void
STImage::Draw() {

	glRasterPos2f(0.0f, 0.0f);
	glDrawPixels(mWidth, mHeight, GL_RGBA, GL_UNSIGNED_BYTE, (GLvoid*)mPixels);

}

