
#include <stdio.h>
#include <math.h>

#include "stgl.h"
#include "STShape.h"


/**
* Generates a circle shape with radius 'r'.  The center of
* the circle is at the origin.
*/
STShape*
STShape::CreateCircle(float r) {

	STShape* sh = new STShape();
	
	// hardcoded number of samples for now
	int samples = 256;

	STPoint2D pt;

	sh->mFaceVertIndices.resize(3*samples);
	sh->mVerts.resize(samples+1);

	pt.x = pt.y = 0.0f;
	sh->mVerts[0] = pt;

	for (int i=0; i<samples; i++) {
		float theta = 2.0f * M_PI * (float)i / (float)samples;
		pt.x = r * cos(theta);
		pt.y = r * sin(theta);
		sh->mVerts[i+1] = pt;
	}

	int idx = 0;
	for (int i=1; i<samples; i++) {
		sh->mFaceVertIndices[idx++] = 0;
		sh->mFaceVertIndices[idx++] = i;
		sh->mFaceVertIndices[idx++] = i+1;
	}

	sh->mFaceVertIndices[idx++] = 0;
	sh->mFaceVertIndices[idx++] = samples-1;
	sh->mFaceVertIndices[idx++] = 1;


	return sh;
}

/**
* Generates a rectnagle shape with width 'w' and height 'h'.  The center of
* the rectangle is at the origin.
*/

STShape*
STShape::CreateRect(float w, float h) {

	STShape* sh = new STShape();

	sh->mVerts.resize(4);
	sh->mFaceVertIndices.resize(6);

	float halfw = w/2.0f;
	float halfh = h/2.0f;


	sh->mVerts[0] = STPoint2D(-halfw, -halfh);
	sh->mVerts[1] = STPoint2D(halfw, -halfh);
	sh->mVerts[2] = STPoint2D(halfw, halfh);
	sh->mVerts[3] = STPoint2D(-halfw, halfh);

	sh->mFaceVertIndices[0] = 0;
	sh->mFaceVertIndices[1] = 1;
	sh->mFaceVertIndices[2] = 2;

	sh->mFaceVertIndices[3] = 0;
	sh->mFaceVertIndices[4] = 2;
	sh->mFaceVertIndices[5] = 3;

	return sh;
}


STShape::STShape() {
	
}


STShape::~STShape() {

}

/**
* Initializes the shape object by parsing the shape file given by 'filename'
*/
STStatus
STShape::Load(const char* filename) {

	// Loads a file format that is very similar to OBJ
	//
	// v   x y [z]
	// vn  x y z
	// vt  x y [z]
	// f   i j k

	FILE* shapeFile = fopen(filename, "r");

	if (!shapeFile) {
		fprintf(stderr, "Could not open shape file '%s'.\n", filename);
		return ST_ERROR;
	}

	char line[80];
	int currentLine = 1;

	while (fgets(line, 80, shapeFile)) {
		
		char* str = strtok(line, " \t\n\r");
		STPoint2D pt;

		if (!str) {

			//printf("Empty line.\n");

		// vertex line.  Read the vertex data (x,y)
		} else if (strcmp(str, "v") == 0) {
	
			str = strtok(NULL, "");
			sscanf(str, "%f %f\n", &pt.x, &pt.y);
			mVerts.push_back( pt );

			//printf("Vertex %f %f\n", pt.x, pt.y);

		// vertex texture coordinate line.  Read the texture coord data
		} else if (strcmp(str, "vt") == 0) {

			str = strtok(NULL, "");
			sscanf(str, "%f %f\n", &pt.x, &pt.y);
			mTexCoords.push_back( pt );

			//printf("TexCoord %f %f\n", pt.x, pt.y);

		// triangle face line.  Will contain three vertex indices
		// (potentially per-vertex texcoords as well)
		// will be of the format i j k
		// or i/ti j/tj k/tk
		} else if (strcmp(str, "f") == 0) {

			int idx[3];

			if (strchr(str, '/') == NULL) {

				// There are no per-vertex texture coordinates specified

				str = strtok(NULL, "");
				//printf("Scanning '%s'\n", str);

				sscanf(str, "%d %d %d\n", &idx[0], &idx[1], &idx[2]);
				mFaceVertIndices.push_back(idx[0]-1);
				mFaceVertIndices.push_back(idx[1]-1);
				mFaceVertIndices.push_back(idx[2]-1);
			
				//printf("Face %d %d %d\n", idx[0], idx[1], idx[2]);

			} else {

				// per-vertex texture coordinates specified

				sscanf(str, "%d/%d", &idx[0], &idx[1]);
				mFaceVertIndices.push_back(idx[0]-1);
				mFaceTexIndices.push_back(idx[1]-1);

				//printf("Face %d/%d ", idx[0], idx[1]);

				sscanf(str, "%d/%d", &idx[0], &idx[1]);
				mFaceVertIndices.push_back(idx[0]-1);
				mFaceTexIndices.push_back(idx[1]-1);

				//printf("%d/%d ", idx[0], idx[1]);

				sscanf(str, "%d/%d", &idx[0], &idx[1]);
				mFaceVertIndices.push_back(idx[0]-1);
				mFaceTexIndices.push_back(idx[1]-1);

				//printf("%d/%d\n", idx[0], idx[1]);
			}

		} else if (str[0] == '#' || strcmp(str, "") == 0 || strcmp(str, "vn") == 0) {

			// do nothing, skip the line.
			//printf("Skipped line: '%s'\n", str);

		// unexpected. Fail out
		} else {
			fprintf(stderr, "Error reading shape data on line %d\n", currentLine);
			fprintf(stderr, "First token was: '%s'\n", str);
			fclose(shapeFile);
			return ST_ERROR;
		}

		fflush(stdout);
		currentLine++;
	}

	fclose(shapeFile);

	/*
	printf("there are %d vertices.\n", mVerts.size());
	printf("there are %d texcoords.\n", mTexCoords.size());
	printf("there are %d faces, %d has tex coords.\n", mFaceVertIndices.size()/3, mFaceTexIndices.size()/3);

	for (size_t i=0; i<mVerts.size(); i++)
		printf("%f %f\n", mVerts[i].x, mVerts[i].y);

	for (size_t i=0; i<mFaceVertIndices.size(); )
		printf("%d %d %d\n", mFaceVertIndices[i++], mFaceVertIndices[i++], mFaceVertIndices[i++]);
	*/

	return ST_OK;
}


/**
* Returns the color of the shape
*/
STColor
STShape::GetColor() const {
	return mColor;
}

/**
* Sets the color of the shape to 'c'
*/
void
STShape::SetColor(const STColor& c) {
	mColor = c;
}

/**
* Draws the shape using GL_TRIANGLES.  In the future this should be
* optimized to use OpenGL display lists since our internal shape
* representation anticipates this.
*/
void
STShape::Draw() {

	glColor3f(mColor.r, mColor.g, mColor.b);
	glBegin(GL_TRIANGLES);

	int numFaces = (int)mFaceVertIndices.size() / 3;

	if (mFaceTexIndices.size() > 0) {
	
		// draw with texture coordinates

		int j = 0;

		for (int i=0; i<numFaces; i++) {

			int vidx, tidx;

			vidx = mFaceVertIndices[j];
			tidx = mFaceTexIndices[j++];
			glTexCoord2f(mTexCoords[tidx].x, mTexCoords[tidx].x);
			glVertex2f(mVerts[vidx].x, mVerts[vidx].y);

			vidx = mFaceVertIndices[j];
			tidx = mFaceTexIndices[j++];
			glTexCoord2f(mTexCoords[tidx].x, mTexCoords[tidx].x);
			glVertex2f(mVerts[vidx].x, mVerts[vidx].y);

			vidx = mFaceVertIndices[j];
			tidx = mFaceTexIndices[j++];
			glTexCoord2f(mTexCoords[tidx].x, mTexCoords[tidx].x);
			glVertex2f(mVerts[vidx].x, mVerts[vidx].y);
		}

	} else {

		// draw without tex coords

		int j = 0;

		for (int i=0; i<numFaces; i++) {

			int idx = mFaceVertIndices[j++];

			glVertex2f(mVerts[idx].x, mVerts[idx].y);
			idx = mFaceVertIndices[j++];
			glVertex2f(mVerts[idx].x, mVerts[idx].y);
			idx = mFaceVertIndices[j++];
			glVertex2f(mVerts[idx].x, mVerts[idx].y);
		}
	}

	glEnd();
}


